#include <mpi.h>
#include <vector>
#include <bitset>
#include <iterator>
#include <algorithm>
#include "defs.h"
#include "alloc.h"
#include "timer.h"


template <typename T>
void do_intermix_sequence(std::vector<T>& sequence, int A) {
    int N = sequence.size();
    if ((N & (N - 1)) != 0) {
        throw std::runtime_error("do_intermix_sequence: N must be a power of 2.");
    }
    A = A % N;
    // Convert A to binary and pad with zeros to match the length of the sequence
    std::bitset<64> binary_A(A);
    std::string binary_str = binary_A.to_string().substr(64 - N);
    // Apply the intermixing based on the binary representation of A
    for (int i = 0; i < N; ++i) {
        // Determine the bit at position i in the binary representation of A
        int bit = binary_str[i] - '0';
        // If bit is 1, perform a swap
        if (bit) {
            // Swap elements at positions i and N-i-1
            std::swap(sequence[i], sequence[N-i-1]);
        }
    }
}


void peer2peer_exchange(const std::vector<int> &params, olocal &output) {
    if (params.size() != 6) {
        throw std::runtime_error("peer2peer_exchange: wrong parameters");
    }
    int rank = params[0];
    int size = params[1];
    int intermix = params[2];
    int npeers_per_node = params[3];
    bool use_gpu_memory = (bool)params[4];
    bool bidirectional = (bool)params[5];
    bool is_intra_node = (size <= npeers_per_node);
    std::string kind = (is_intra_node ? "intra-node" : "cross-node");

    if (intermix >= size/2) {
        throw std::runtime_error("peer2peer_exchange: wrong parameters");
    }

    std::vector<int> intermixed_sequence;
    for (int i = 0; i < size/2; ++i) {
        intermixed_sequence.push_back(i);
    }
    do_intermix_sequence(intermixed_sequence, intermix);

    bool do_recv_first = false;
    int peer;
	int internal_peer;
    int internal_rank = rank % size;
    if (internal_rank >= size/2) {
        do_recv_first = true;
        internal_peer = (int)std::distance(intermixed_sequence.begin(), std::find(intermixed_sequence.begin(), intermixed_sequence.end(), internal_rank - size/2));
    } else {
        internal_peer = intermixed_sequence[internal_rank] + size/2;    
    }
    peer = internal_peer + size * (rank/size);
    MPI_Comm comm;
	if (is_intra_node) {
#if WITH_PER_NODE_COMM
	    MPI_CALL(MPI_Comm_split_type(MPI_COMM_WORLD, MPI_COMM_TYPE_SHARED, 0, MPI_INFO_NULL, &comm));
        peer = peer % npeers_per_node;
#else
		MPI_CALL(MPI_Comm_dup(MPI_COMM_WORLD, &comm));
#endif
	} else {
		MPI_CALL(MPI_Comm_dup(MPI_COMM_WORLD, &comm));
	}
    if (peer >= size || peer < 0) {
	    MPI_CALL(MPI_Comm_free(&comm));
        return;  
    }
    int niters = 50;
    int nwarmup = 3;
    size_t bufsize = (1<<23) * sizeof(real);
    size_t cachesize = 1<<22;
    char *buf = nullptr;
    if (use_gpu_memory) {
        buf = sys::gpu_alloc(bufsize + cachesize);
    } else {
#if PINNED_FOR_MPI_TRANSFERS && WITH_GPU
        buf = sys::cpu_alloc_pinned(bufsize + cachesize);
#else
        buf = sys::cpu_alloc_pageable(bufsize + cachesize);
#endif
    }
    std::vector<double> result;
	if (!bidirectional) {
        timer t(rank2str(rank) + "Peer-to-peer bandwidth unidirectional", bufsize, niters, result);
        t.comment << kind << " with intermix=" << intermix;
        for (int i = 0; i < niters + nwarmup; i++) {
            if (i == nwarmup)
                t.start();
            if (do_recv_first) {
                MPI_CALL(MPI_Recv(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, MPI_STATUS_IGNORE));
            } else {
                MPI_CALL(MPI_Send(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm));
            }
        }
    } else {
        timer t(rank2str(rank) + "Peer-to-peer bandwidth bidirectional", bufsize, niters, result);
        t.comment << kind << " with intermix=" << intermix;
		MPI_Request request;
        for (int i = 0; i < niters + nwarmup; i++) {
            if (i == nwarmup)
                t.start();
			MPI_CALL(MPI_Isend(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, &request));
			MPI_CALL(MPI_Recv(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, MPI_STATUS_IGNORE));
			MPI_CALL(MPI_Wait(&request, MPI_STATUS_IGNORE));
        }
    }
    if (use_gpu_memory) {
        sys::gpu_free(buf);
    } else {
#if PINNED_FOR_MPI_TRANSFERS && WITH_GPU
        sys::cpu_free_pinned(buf);
#else
        sys::cpu_free_pageable(buf);
#endif
    }
	MPI_CALL(MPI_Comm_free(&comm));
    output.values.push_back(result);
    output.header = std::string("MPI_P2P_") + (is_intra_node ? "intranode_" : "crossnode_") + (use_gpu_memory ? "GPU_" : "CPU_") + (bidirectional ? "2DIR" : "1DIR");
}

